# -*- coding: UTF-8 -*-
import asyncio
import os
import random
import threading
import time
from typing import Callable

from common import logger

log = logger.getUnloadLogger('main')


def worker(task_id: str, func: Callable, kwargs):
    """Выполнение задачи - вызов функции из unloadService"""

    log.debug(f"Task {task_id} - Thread: {threading.get_ident()}, Process: {os.getpid()}")
    try:
        res = func(**kwargs)
        time.sleep(random.uniform(0.1, 0.5))
        return res
    except asyncio.exceptions.CancelledError as e:
        log.debug(e)
        return False, task_id, str(e)


class AsyncTaskManager:
    def __init__(self, max_workers: int = 32, max_concurrent_tasks: int = 100, task_await_timeout=None):
        self.max_concurrent_tasks = max_concurrent_tasks if max_concurrent_tasks else 100
        self.max_workers = max_workers if max_workers else 32
        self.task_await_timeout = task_await_timeout if task_await_timeout else 10

        self.queue = asyncio.Queue()
        self.pending_tasks = set()

        self.all_tasks = set()
        self.processed_tasks = set()

    async def add_task(self, task_id, task_func: Callable, done_callback=Callable, **kwargs):
        """Добавление задачи в очередь"""
        await self.queue.put((task_id, task_func, done_callback, kwargs))
        return task_id

    def task_done_callback(self, future, task_id, custom_callback: Callable):
        log.debug(f"Завершена задача {task_id} - thread: {threading.get_ident()}, pid: {os.getpid()}")
        try:
            self.pending_tasks.remove(future)
            self.processed_tasks.add(task_id)
            custom_callback(future)
        except asyncio.CancelledError as e:
            log.debug(f"Task {task_id} cancelled {e}")

    async def process_tasks(self):
        """Обработка очереди задач"""
        while True:
            """
            собираем задачи из очереди, чтобы не обрабатывать по одной
            порядок работы: 
            0. из каждой задачи в очереди создаем корутину - асинхронную задачу, которая выполняется в отдельном потоке
            1. собираем задачи из очереди в список - накапливаем кол-во = max_concurrent_tasks, чтобы начать обработку.
            если queue_size < max_concurrent_tasks, то берем все задачи из очереди
            2. при выполнении задачи распределяются между воркерами(max_workers)
            3. для каждой задачи асинхронно вызывается worker(), где выполняется функция выгрузки продаж
            4. для каждой задачи асинхронно вызывается done_callback(), где выполняется функция постобработки
            """

            if self.queue.empty() and len(self.pending_tasks) == 0:
                log.debug("В очереди пусто, все задачи выполнены")
                break
            elif self.queue.empty() and 0 < len(self.pending_tasks) < self.max_concurrent_tasks:
                log.debug(f"Выполняем {len(self.pending_tasks)} задач")
                await asyncio.gather(*self.pending_tasks, return_exceptions=False)
                continue
            if len(self.pending_tasks) >= self.max_concurrent_tasks:
                tasks = set(list(self.pending_tasks)[:self.max_concurrent_tasks])
                log.debug(f"Выполняем {len(tasks)} задач")
                await asyncio.gather(*tasks, return_exceptions=False)
                continue

            # next task from queue
            # task_id = cash_id
            task_id, task_func, done_callback, kwargs = await self.queue.get()
            # работаем с циклом событий, чтобы установить туда макс. кол-во порождаемых процессов
            # по умолчанию  min(32, os.cpu_count() + 4)
            # loop = asyncio.get_running_loop()
            # loop.set_default_executor(ThreadPoolExecutor(self.max_workers))
            # задачу на выгрузку превращаем в future
            loop = asyncio.get_event_loop()
            coro = loop.run_in_executor(None, worker, task_id, task_func, kwargs)
            coro.add_done_callback(lambda f: self.task_done_callback(f, task_id, done_callback))
            self.pending_tasks.add(coro)
            # информационный вывод статуса очереди - на всякий случай
            print(f"\nCurrent queue status: \n- "
                  f"Tasks in queue: {self.queue.qsize()} \n- Pending tasks: {len(self.pending_tasks)} \n- Last added task ID: {task_id}")

    async def run_until_complete(self):
        processor_task = asyncio.create_task(self.process_tasks())
        try:
            # оставляем какое-время на обработку всех задач из очереди
            # застрявшие в очереди задачи закрываем во избежание заморозки процессов в системе
            # в норме к этому моменту очередь уже будет пуста, просто ждем еще на всякий случай несколько секунд
            await asyncio.wait_for(self.queue.join(), timeout=20)
        except asyncio.TimeoutError:
            log.debug("Заканчиваем процесс выполнения задач...")
        finally:
            processor_task.cancel()
            asyncio.get_running_loop().close()
